#include <stdio.h>
#include "opencv2/opencv.hpp"

#define PIXVAL(iplimagep, x, y) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x)))

#define MAX_FEAT 1024

int main()
{
    int i;
    int count;
    int key;
    int track_mode = 0; /* 0: optical flow mode, 1: tracking mode */
    int blackout = 0; /* 0: show image, 1: blackout */
    int init_flag = 0;
    char status[MAX_FEAT];
    double qual_level = 0.005, min_dist = 5.0;
    IplImage *frame, *buf, *curr, *prev;
    IplImage *eig, *tmp, *pyr_curr, *pyr_prev;
    CvPoint2D32f *feat_curr, *feat_prev, *feat_tmp_ptr;
    CvTermCriteria criteria;
    CvCapture *capture = cvCaptureFromCAM(0);
    if (capture == NULL) {
        fprintf(stderr, "cannot capture\n");
        return 1;
    }

    frame = cvQueryFrame(capture);
    buf = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);  // color
    curr = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);  // mono
    prev = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1); // previous frame

    eig = cvCreateImage(cvGetSize(frame), IPL_DEPTH_32F, 1);
    tmp = cvCreateImage(cvGetSize(frame), IPL_DEPTH_32F, 1);
    pyr_curr = cvCreateImage(cvSize(frame->width + 8, frame->height / 3),
                             IPL_DEPTH_8U, 1);
    pyr_prev = cvCreateImage(cvSize(frame->width + 8, frame->height / 3),
                             IPL_DEPTH_8U, 1);

    feat_curr = (CvPoint2D32f *)malloc(MAX_FEAT * sizeof(CvPoint2D32f));
    feat_prev = (CvPoint2D32f *)malloc(MAX_FEAT * sizeof(CvPoint2D32f));
    if (feat_curr == NULL || feat_prev == NULL) {
        fprintf(stderr, "cannot malloc feat_curr or feat_prev\n");
        return 1;
    }

    criteria = cvTermCriteria(CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 5, 0.03);

    cvNamedWindow("image", CV_WINDOW_AUTOSIZE);
    printf("q: quit, t: toggle flow/track, b: toggle blackout, i: init feat. points\n");

    while (1) {
        frame = cvQueryFrame(capture); /* cannot be modified */

        cvCopy(frame, buf);
        cvCvtColor(buf, curr, CV_BGR2GRAY);

        if (!track_mode || init_flag) {
            count = MAX_FEAT;
            cvGoodFeaturesToTrack(prev, eig, tmp, feat_prev, &count,
                                  qual_level, min_dist
                                  /*, NULL, 3, 1, 0.04 */);
                  /* uncomment the last line to use the Harris operator */
            init_flag = 0;
        } else {
            feat_tmp_ptr = feat_prev;
            feat_prev = feat_curr;
            feat_curr = feat_tmp_ptr;
        }

        cvCalcOpticalFlowPyrLK(prev, curr, pyr_prev, pyr_curr,
                               feat_prev, feat_curr, count, cvSize(10, 10),
                               3, status, NULL, criteria, 0);
        cvCopy(curr, prev, 0);

        if (blackout) {
            cvZero(buf);
        } else {
            cvCvtColor(curr, buf, CV_GRAY2BGR);
        }

        for (i = 0; i < count; i++) {
            if (status[i] == 0) { continue; }
            if (track_mode == 0) {
                cvLine(buf, cvPointFrom32f(feat_prev[i]),
                       cvPointFrom32f(feat_curr[i]),
                       CV_RGB(0, 255, 0));
            } else {
                cvCircle(buf, cvPointFrom32f(feat_curr[i]), 2,
                         CV_RGB(0, 255, 0), -1);
            }
        }

        cvShowImage("image", buf);

        if ((key = cvWaitKey(30)) > 0) {
            if (key == 'q' || key == 0x1b) {
                break;
            } else if (key == 't') {
                track_mode = 1 - track_mode;
            } else if (key == 'b') {
                blackout = 1 - blackout;
            } else if (key == 'i') {
                init_flag = 1;
            } else if (key == 'c') {
                cvSaveImage("saved0.jpg", buf);
            }
        }
    }

    return 0;
}
